import 'dart:async';
import 'dart:typed_data';

// import 'package:fis_measure/view/frame_view/native/frame_view.dart';
import 'package:fis_common/logger/logger.dart';
import 'package:fis_measure/view/vid_painter/vid_painter.dart';
import 'package:flutter/material.dart' hide Image, decodeImageFromList;
import 'dart:ui';

import 'controller.dart';
import 'enums.dart';
import 'events.dart';

class VidPlayer extends StatefulWidget {
  const VidPlayer(
    this.controller, {
    Key? key,
    this.width,
    this.height,
  }) : super(key: key);

  final VidPlayerController controller;

  final double? width;
  final double? height;

  @override
  State<StatefulWidget> createState() => _VidPlayerState();
}

class _VidPlayerState extends State<VidPlayer> {
  Uint8List? frameBytes;
  Image? image;
  int frameIndex = 0;
  List<double> curColorFilterMatrix = <double>[
    1, 0, 0, 0, 0, // red
    0, 1, 0, 0, 0, // green
    0, 0, 1, 0, 0, // blue
    0, 0, 0, 1, 0, // alpha
  ];
  @override
  void initState() {
    widget.controller.eventHandler.addListener(onControllerEvent);
    super.initState();
  }

  @override
  void didUpdateWidget(covariant VidPlayer oldWidget) {
    if (oldWidget.controller != widget.controller) {
      throw UnsupportedError("[VidTestPlayer] unsupport replace controller.");
    }
    super.didUpdateWidget(oldWidget);
  }

  void onControllerEvent(Object sender, VidPlayerEvent e) {
    if (e is VidPlayerStatusChangeEvent) {
      if (!widget.controller.hasView) {
        setState(() {});
      }
    }
    if (e is VidPlayerFrameIndexChangeEvent) {
      onFrameChanged(e);
    }
    if (e is VidPlayerFilterChangeEvent) {
      onFilterMatrixChanged(e);
    }
  }

  void onFrameChanged(VidPlayerFrameIndexChangeEvent e) {
    loadFrame(e);
  }

  void onFilterMatrixChanged(VidPlayerFilterChangeEvent e) {
    setState(() {
      curColorFilterMatrix = e.filterMatrix;
    });
  }

  void loadFrame(VidPlayerFrameIndexChangeEvent e) async {
    frameIndex = e.index;
    final oldImage = image;
    image = await _loadImage(e.bytes);
    try {
      oldImage?.dispose();
    } catch (e) {
      logger.e("Old image dispose error in vidplayer", e);
    }
    setState(() {});
  }

  void updateFrame() {
    setState(() {});
  }

  Future<Image> _loadImage(Uint8List bytes) async {
    final Completer<Image> completer = Completer<Image>();
    decodeVidFromList(bytes, (Image image) => completer.complete(image));
    // decodeImageFromList(bytes, (Image image) => completer.complete(image));
    return completer.future;
  }

  void decodeVidFromList(Uint8List bytes, ImageDecoderCallback callback) async {
    ImmutableBuffer.fromUint8List(bytes).then((ImmutableBuffer buffer) async {
      final ImageDescriptor descriptor = await ImageDescriptor.encoded(buffer);
      descriptor.instantiateCodec().then((Codec codec) {
        final Future<FrameInfo> frameInfo = codec.getNextFrame();
        codec.dispose();
        return frameInfo;
      }).then((FrameInfo frameInfo) {
        buffer.dispose();
        descriptor.dispose();
        return callback(frameInfo.image);
      });
    });
  }

  @override
  Widget build(BuildContext context) {
    Widget? child;
    switch (widget.controller.status) {
      case VidPlayStatus.init:
        child = Container(child: const Text("Loading"));
        break;
      case VidPlayStatus.ready:
        child = Container(child: const Text("Ready"));
        break;
      case VidPlayStatus.loadFail:
        child = Container(child: const Text("Load fail"));
        break;
      case VidPlayStatus.play:
      case VidPlayStatus.pause:
        child = buildFrameView(context);
        break;
      case VidPlayStatus.stop:
      case VidPlayStatus.dispose:
        child = const Text("Closed");
        break;
    }
    return buildBox(context, child);
  }

  Widget buildBox(BuildContext context, Widget child) {
    return Container(
      alignment: Alignment.center,
      child: child,
    );
  }

  Widget buildFrameView(BuildContext context) {
    if (image != null) {
      final size = MediaQuery.of(context).size;
      return RepaintBoundary(
        child: CustomPaint(
          painter: VidPainter(
            image: image!,
            frameIndex: frameIndex,
            colorFilterMatrix: curColorFilterMatrix,
          ),
          isComplex: false, //是否为复杂图像（true会缓存）
          size: size,
        ),
      );
    } else {
      return Container();
    }
  }

  @override
  void dispose() {
    widget.controller.eventHandler.removeListener(onControllerEvent);
    super.dispose();
  }
}
